# SPDX-License-Identifier: CECILL-C

"""
    struct SpeechDatasetInfos
Store metadata about a dataset.
# Fields
- `name` Dataset official name
- `lang` Language or list of languages (ISO 639-3 code)
- `license` License name
- `source` URL to the dataset publication or content
- `authors` list of authors
- `description` A few sentences describing the content or main purpose 
- `subsets` List of available subsets (for example ["train", "test"])
"""
@kwdef struct SpeechDatasetInfos
    name::AbstractString = ""
    lang::Union{AbstractString, Vector{AbstractString}} = ""
    license::AbstractString = ""
    source::AbstractString = ""
    authors::Vector{AbstractString} = []
    description::AbstractString = ""
    subsets::Vector{AbstractString} = []
end

function SpeechDatasetInfos(infos::AbstractDict)
    kwargs = NamedTuple()
    for key in fieldnames(SpeechDatasetInfos)
        val = get(infos, String(key), nothing)
        # merge new (key=val) if key was found
        kwargs = !isnothing(val) ? (; kwargs..., key=>val) : kwargs
    end
    SpeechDatasetInfos(;kwargs...)
end

"""
    SpeechDatasetInfos(name::AbstractString)
Construct a SpeechDatasetInfos from the Dataset name.
"""
function SpeechDatasetInfos(name::AbstractString)
    corpora_infos = JSON.parsefile(corpora_file)
    infos = filter(x -> x["name"]==name, corpora_infos)[1]
    SpeechDatasetInfos(infos)
end

"""
    struct SpeechDataset <: MLUtils.AbstractDataContainer
Store all dataset recordings and annotations. \n
It can be iterated, and will give a `Tuple{Recording, Annotation}` for each entry. Indexation can be done with integer or id.
# Fields
- `infos::SpeechDatasetInfos`
- `idxs::Vector{AbstractString}` id indexes to access elements
- `annotations::Dict{AbstractString, Annotation}` Annotation for each index
- `recordings::Dict{AbstractString, Recording}` Recording for each index
"""
struct SpeechDataset <: MLUtils.AbstractDataContainer
    infos::SpeechDatasetInfos
    idxs::Vector{AbstractString}
    annotations::Dict{AbstractString, Annotation}
    recordings::Dict{AbstractString, Recording}
end

function SpeechDataset(infos::SpeechDatasetInfos, annotations::Dict{AbstractString, Annotation}, recordings::Dict{AbstractString, Recording})
    idxs = collect(keys(annotations))
    SpeechDataset(infos, idxs, annotations, recordings)
end

"""
    SpeechDataset(infos::SpeechDatasetInfos, manifestroot::AbstractString, subset::AbstractString)
Create a SpeechDataset from manifest files and subset.
"""
function SpeechDataset(infos::SpeechDatasetInfos, manifestroot::AbstractString, subset::AbstractString)
    subset_name = subset == "" ? "" : "-$(subset)"
    annot_path =  joinpath(manifestroot, "annotations$(subset_name).jsonl") 
    rec_path = joinpath(manifestroot, "recordings.jsonl")
    annotations = load_manifest(Annotation, annot_path)
    recordings = load_manifest(Recording, rec_path)
    SpeechDataset(infos, annotations, recordings)
end

Base.getindex(d::SpeechDataset, key::AbstractString) = d.recordings[key], d.annotations[key]
Base.getindex(d::SpeechDataset, idx::Integer) = getindex(d, d.idxs[idx])
# Fix1 -> partial function with fixed 1st argument
Base.getindex(d::SpeechDataset, idxs::AbstractVector) = map(Base.Fix1(getindex, d), idxs)

Base.length(d::SpeechDataset) = length(d.idxs)

function Base.filter(fn, d::SpeechDataset)
    fidxs = filter(d.idxs) do i
        fn((d.recordings[i], d.annotations[i]))
    end
    idset = Set(fidxs)

    fannotations = filter(d.annotations) do (k, v)
        k ∈ idset
    end

    frecs = filter(d.recordings) do (k, v)
        k ∈ idset
    end

    SpeechDataset(d.infos, fidxs, fannotations, frecs)
end

"""
    struct_summary(object; additional=[])
Display fields and values of given object.
Can display additional informations if provided as (n,2) matrix
"""
function struct_summary(object; additional=[])
    properties = collect(propertynames(object))
	values = [getproperty(object, p) for p in properties]
    description = Array{String}(undef, length(properties), 2)
	description[:,1] = string.(properties)
	description[:,2] = string.(values)
    if ! isempty(additional)
        description = vcat(description, additional)
        push!(values, additional[:,1]...)
    end
    gap = maximum(length.(description[:,1]))+4
    for (i, line) in enumerate(eachrow(description))
        # remove vector type from string representation
        if typeof(values[i]) <: Vector
            line[2] = replace(line[2], r".*\["=>"[")
        end
        # print <field> <gap> <value>
        println(line[1], repeat(' ', gap-length(line[1])), line[2])
    end
end

"""
    Base.summary(infos::SpeechDatasetInfos)
Display fields and values of given SpeechDatasetInfos
"""
function Base.summary(infos::SpeechDatasetInfos)
    struct_summary(infos)
end

"""
    Base.summary(dataset::SpeechDataset)
Display informations about given SpeechDataset
"""
function Base.summary(dataset::SpeechDataset)
    additional = ["elements" string(length(dataset))]
    struct_summary(dataset.infos, additional=additional)
end

function get_outfiles(subsets)
    outfiles = ["recordings.jsonl"]
    if isempty(subsets)
        push!(outfiles, "annotations.jsonl")
    else 
        annotationsfiles = ["annotations-$subset.jsonl" for subset in subsets]
        push!(outfiles, annotationsfiles...)
    end
    outfiles
end

"""
    dataset(name::AbstractString, inputdir::AbstractString, outputdir::AbstractString; <keyword arguments>)
Extract recordings and annotations for desired dataset.\n
Return a SpeechDataset object.\n
Create the `outputdir` folder, with:
- `recordings.jsonl` containing each audio file path and associated metadata
- `annotations-<subset>.jsonl` containing each annotation and associated metadata
# Arguments
- `name` Name of the dataset. Supported names are $corpora_names.
- `inputdir` Name of dataset directory. If the directory does not exists, it is created and the data is downloaded if possible. Not all datasets can be downloaded, for example proprietary datasets does not implements a download function.
- `outputdir` is the output directory for manifest files.
# Keyword Arguments
Common kwargs are
- `subset` Part of the dataset to load (for example "train" or "test").
- `lang` ISO 639-3 code of the language.
Other kwargs can be available depending on the dataset, they can be accessed with [`get_dataset_kwargs(name::String)`](@ref).
"""
function dataset(name::AbstractString, inputdir::AbstractString, outputdir::AbstractString; kwargs...)
    # check name
    name ∈ corpora_names || throw(ArgumentError("Name $name is not supported, try one of $corpora_names."))
    nametype = get_nametype(name)
    dataset(DatasetBuilder(nametype), name, inputdir, outputdir; kwargs...)
end

function dataset(builder::DatasetBuilder, name::AbstractString, inputdir::AbstractString, outputdir::AbstractString; subset="", kwargs...)
    kwargs = values(kwargs)
    # check kwargs
    all([k in keys(builder.kwargs) for k in keys(kwargs)]) || throw(ArgumentError("Unsupported keyword argument. Available are $(builder.kwargs)"))

    # download if possible and inputdir doesn't exists yet  
    downloadable = hasmethod(download, Tuple{typeof(builder), AbstractString})
    if downloadable && ! isdir(inputdir)
        download(builder, inputdir)
    end

    # load infos from dataset name
    infos = SpeechDatasetInfos(name)

    # Check subset value 
    subset ∈ [infos.subsets ; ""] || throw(ArgumentError("Subset $subset is not supported, try one of $(infos.subsets)."))

    # Check lang value if provided
    if :lang ∈ keys(kwargs) 
        kwargs.lang ∈ infos.lang || throw(ArgumentError("Language $lang is not supported, try one of $(infos.lang)."))
    end

    # prepare if outfiles are not present
    outfiles = get_outfiles(infos.subsets)
    if ! all(isfile.(joinpath.(outputdir, outfiles)))
        prepare(builder, inputdir, outputdir; kwargs...)
    end

    SpeechDataset(infos, outputdir, subset)
end
